package de.hft.stuttgart.citydoctor2.datastructure;

import de.hft.stuttgart.citydoctor2.check.Check;
import de.hft.stuttgart.citydoctor2.check.CheckError;
import de.hft.stuttgart.citydoctor2.check.CheckId;
import de.hft.stuttgart.citydoctor2.parser.ParserConfiguration;
import de.hft.stuttgart.citydoctor2.utils.CityGmlUtils;
import de.hft.stuttgart.citydoctor2.utils.CopyHandler;
import de.hft.stuttgart.citydoctor2.utils.Copyable;
import org.apache.logging.log4j.LogManager;
import org.apache.logging.log4j.Logger;
import org.citygml4j.core.model.core.AbstractCityObject;
import org.citygml4j.core.model.core.AbstractSpaceBoundaryProperty;
import org.citygml4j.core.util.geometry.GeometryFactory;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurface;
import org.xmlobjects.gml.model.geometry.aggregates.MultiSurfaceProperty;
import org.xmlobjects.gml.model.geometry.primitives.Solid;
import org.xmlobjects.gml.model.geometry.primitives.SolidProperty;

import java.io.Serial;
import java.util.ArrayList;
import java.util.List;

/**
 *
 */
public class GenericCityObject extends CityObject{

    @Serial
    private static final long serialVersionUID = 5886527664552294353L;

    private org.citygml4j.core.model.generics.GenericOccupiedSpace cgmlGos;

    private static final Logger logger = LogManager.getLogger(GenericCityObject.class);

    private final List<BoundarySurface> boundarySurfaceList = new ArrayList<>();

    @Override
    public void accept(Check c) {
        super.accept(c);
        if (c.canExecute(this)) {
            c.check(this);
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.accept(c);
        }
    }

    @Override
    public void reCreateGeometries(GeometryFactory factory, ParserConfiguration config) {
        for (Geometry geom : getGeometries()) {
            if (geom.getType() == GeometryType.MULTI_SURFACE) {
                MultiSurface ms = CityGmlUtils.createMultiSurface(geom, factory, config);
                setMultiSurfaceAccordingToLod(geom, ms);
            } else {
                Solid solid = CityGmlUtils.createSolid(geom, factory, config);
                setSolidAccordingToLod(geom, solid);
            }
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            reCreateBoundarySurface(factory, config, bs);
        }
    }

    private void reCreateBoundarySurface(GeometryFactory factory, ParserConfiguration config, BoundarySurface bs) {
        if (bs.getGeometries().isEmpty()) {
            for (AbstractSpaceBoundaryProperty bsp : cgmlGos.getBoundaries()) {
                if (bsp.getObject() != null && bsp.getObject() == bs.getGmlObject()) {
                    logger.warn("Found empty boundary surface: {}, removing from generic object", bs.getGmlId());
                    cgmlGos.getBoundaries().remove(bsp);
                    break;
                }
            }
            return;
        }
        bs.reCreateGeometries(factory, config);
    }

    private void setMultiSurfaceAccordingToLod(Geometry geom, MultiSurface ms) {
        switch (geom.getLod()) {
            case LOD0:
                cgmlGos.setLod0MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD2:
                cgmlGos.setLod2MultiSurface(new MultiSurfaceProperty(ms));
                break;
            case LOD3:
                cgmlGos.setLod3MultiSurface(new MultiSurfaceProperty(ms));
                break;
            default:
                throw new IllegalStateException("Cannot add " + geom.getLod() + " multi surface to buildings");
        }
    }

    private void setSolidAccordingToLod(Geometry geom, Solid solid) {
        switch (geom.getLod()) {
            case LOD1:
                cgmlGos.setLod1Solid(new SolidProperty(solid));
                break;
            case LOD2:
                cgmlGos.setLod2Solid(new SolidProperty(solid));
                break;
            case LOD3:
                cgmlGos.setLod3Solid(new SolidProperty(solid));
                break;
            default:
                throw new IllegalStateException("Cannot add " + geom.getLod() + " solid to buildings");
        }
    }

    @Override
    public void collectContainedErrors(List<CheckError> errors) {
        super.collectContainedErrors(errors);
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.collectContainedErrors(errors);
        }
    }

    @Override
    public void clearAllContainedCheckResults() {
        super.clearAllContainedCheckResults();
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.clearAllContainedCheckResults();
        }
    }

    @Override
    public boolean containsError(CheckId checkIdentifier) {
        boolean hasError = super.containsError(checkIdentifier);
        if (hasError) {
            return true;
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            if (bs.containsError(checkIdentifier)) {
                return true;
            }
        }
        return false;
    }

    @Override
    public boolean containsAnyError() {
        boolean hasError = super.containsAnyError();
        if (hasError) {
            return true;
        }
        for (BoundarySurface bs : boundarySurfaceList) {
            if (bs.containsAnyError()) {
                return true;
            }
        }
        return false;
    }

    @Override
    public void prepareForChecking() {
        super.prepareForChecking();
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.prepareForChecking();
        }
    }

    @Override
    public void clearMetaInformation() {
        super.clearMetaInformation();
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.clearMetaInformation();
        }
    }

    @Override
    public void collectInstances(CopyHandler handler) {
        super.collectInstances(handler);
        handler.addInstance(boundarySurfaceList);
    }

    @Override
    public void fillValues(Copyable original, CopyHandler handler) {
        super.fillValues(original, handler);
        GenericCityObject originalGos = (GenericCityObject) original;
        for (BoundarySurface originalBs : originalGos.boundarySurfaceList) {
            boundarySurfaceList.add(handler.getCopyInstance(originalBs));
        }
        cgmlGos = originalGos.cgmlGos;
    }

    @Override
    public AbstractCityObject getGmlObject() {
        return null;
    }

    @Override
    public FeatureType getFeatureType() {
        return FeatureType.GENERIC_CITY_OBJECT;
    }

    public List<BoundarySurface> getBoundarySurfaces() {
        return boundarySurfaceList;
    }

    @Override
    public void unsetGmlGeometries() {
        cgmlGos.setLod1Solid(null);
        cgmlGos.setLod2Solid(null);
        cgmlGos.setLod3Solid(null);
        cgmlGos.setLod2MultiSurface(null);
        cgmlGos.setLod3MultiSurface(null);
        for (BoundarySurface bs : boundarySurfaceList) {
            bs.unsetGmlGeometries();
        }
    }

    public void setGmlObject(org.citygml4j.core.model.generics.GenericOccupiedSpace gos){
        this.cgmlGos = gos;
    }

    public void addBoundarySurface(BoundarySurface bs) {
        boundarySurfaceList.add(bs);
        bs.setParent(this);
    }


    @Override
    public Copyable createCopyInstance() {
        return new GenericCityObject();
    }
}
